@@ -10,6 +10,7 @@
  * Bjorn Reese <breese@users.sourceforge.net>
  */
 
+#define IN_LIBXSLT
 #include "libxslt.h"
 
 #include <math.h>
@@ -432,9 +433,9 @@ xsltNumberFormatGetAnyLevel(xsltTransformContextPtr context,
     xsltCompMatchPtr fromPat = NULL;
 
     if (count != NULL)
-	countPat = xsltCompilePattern(count, doc, elem);
+	countPat = xsltCompilePattern(count, doc, elem, NULL, context);
     if (from != NULL)
-	fromPat = xsltCompilePattern(from, doc, elem);
+	fromPat = xsltCompilePattern(from, doc, elem, NULL, context);
 	
     /* select the starting node */
     switch (node->type) {
@@ -516,11 +517,11 @@ xsltNumberFormatGetMultipleLevel(xsltTransformContextPtr context,
     xsltCompMatchPtr fromPat;
 
     if (count != NULL)
-	countPat = xsltCompilePattern(count, doc, elem);
+	countPat = xsltCompilePattern(count, doc, elem, NULL, context);
     else
 	countPat = NULL;
     if (from != NULL)
-	fromPat = xsltCompilePattern(from, doc, elem);
+	fromPat = xsltCompilePattern(from, doc, elem, NULL, context);
     else
 	fromPat = NULL;
     context->xpathCtxt->node = node;
@@ -612,11 +613,13 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
     int amount, i;
     double number;
     xsltFormat tokens;
+    int tempformat = 0;
 
     if ((data->format == NULL) && (data->has_format != 0)) {
 	data->format = xsltEvalAttrValueTemplate(ctxt, data->node,
 					     (const xmlChar *) "format",
 					     XSLT_NAMESPACE);
+	tempformat = 1;
     }
     if (data->format == NULL) {
 	return;
@@ -712,7 +715,12 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
 	    xmlFree(tokens.tokens[i].separator);
     }
     
- XSLT_NUMBER_FORMAT_END:
+XSLT_NUMBER_FORMAT_END:
+    if (tempformat == 1) {
+	/* The format need to be recomputed each time */
+	xmlFree(data->format);
+	data->format = NULL;
+    }
     if (output != NULL)
 	xmlBufferFree(output);
 }
@@ -1085,9 +1093,16 @@ OUTPUT_NUMBER:
     number = fabs(number) * (double)format_info.multiplier;
     scale = pow(10.0, (double)(format_info.frac_digits + format_info.frac_hash));
     number = floor((scale * number + 0.5)) / scale;
-    xsltNumberFormatDecimal(buffer, floor(number), self->zeroDigit[0],
-			    format_info.integer_digits,
-			    format_info.group, (xmlChar)',');
+    if ((self->grouping != NULL) && (self->grouping[0] != 0))
+	xsltNumberFormatDecimal(buffer, floor(number), self->zeroDigit[0],
+				format_info.integer_digits,
+				format_info.group,
+				(xmlChar) self->grouping[0]);
+    else
+	xsltNumberFormatDecimal(buffer, floor(number), self->zeroDigit[0],
+				format_info.integer_digits,
+				format_info.group,
+				(xmlChar) ',');
 
     /* Next the fractional part, if required */
     if (format_info.frac_digits + format_info.frac_hash > 0) {
