@@ -14,6 +14,7 @@
  * TODO: detect [number] at compilation, optimize accordingly
  */
 
+#define IN_LIBXSLT
 #include "libxslt.h"
 
 #include <string.h>
@@ -71,9 +72,9 @@ struct _xsltStepOp {
     /*
      * Optimisations for count
      */
-    xmlNodePtr previous;
-    int        index;
-    int        len;
+    int        previousExtra;
+    int        indexExtra;
+    int        lenExtra;
 };
 
 struct _xsltCompMatch {
@@ -95,6 +96,8 @@ struct _xsltCompMatch {
 typedef struct _xsltParserContext xsltParserContext;
 typedef xsltParserContext *xsltParserContextPtr;
 struct _xsltParserContext {
+    xsltStylesheetPtr style;		/* the stylesheet */
+    xsltTransformContextPtr ctxt;	/* the transformation or NULL */
     const xmlChar *cur;			/* the current char being parsed */
     const xmlChar *base;		/* the full expression */
     xmlDocPtr      doc;			/* the source document */
@@ -189,13 +192,15 @@ xsltFreeCompMatchList(xsltCompMatchPtr comp) {
 
 /**
  * xsltNewParserContext:
+ * @style:  the stylesheet
+ * @ctxt:  the transformation context, if done at run-time
  *
  * Create a new XSLT ParserContext
  *
  * Returns the newly allocated xsltParserContextPtr or NULL in case of error
  */
 static xsltParserContextPtr
-xsltNewParserContext(void) {
+xsltNewParserContext(xsltStylesheetPtr style, xsltTransformContextPtr ctxt) {
     xsltParserContextPtr cur;
 
     cur = (xsltParserContextPtr) xmlMalloc(sizeof(xsltParserContext));
@@ -206,6 +211,8 @@ xsltNewParserContext(void) {
 	return(NULL);
     }
     memset(cur, 0, sizeof(xsltParserContext));
+    cur->style = style;
+    cur->ctxt = ctxt;
     return(cur);
 }
 
@@ -235,19 +242,38 @@ xsltFreeParserContext(xsltParserContextPtr ctxt) {
  * Returns -1 in case of failure, 0 otherwise.
  */
 static int
-xsltCompMatchAdd(xsltCompMatchPtr comp, xsltOp op, xmlChar *value,
-	         xmlChar *value2) {
+xsltCompMatchAdd(xsltParserContextPtr ctxt, xsltCompMatchPtr comp,
+                 xsltOp op, xmlChar * value, xmlChar * value2)
+{
     if (comp->nbStep >= 40) {
-	xsltPrintErrorContext(NULL, NULL, NULL); /* TODO */
+        xsltPrintErrorContext(NULL, NULL, NULL);        /* TODO */
         xsltGenericError(xsltGenericErrorContext,
-		"xsltCompMatchAdd: overflow\n");
-        return(-1);
+                         "xsltCompMatchAdd: overflow\n");
+        return (-1);
     }
     comp->steps[comp->nbStep].op = op;
     comp->steps[comp->nbStep].value = value;
     comp->steps[comp->nbStep].value2 = value2;
+    if (ctxt->ctxt != NULL) {
+	comp->steps[comp->nbStep].previousExtra =
+	    xsltAllocateExtraCtxt(ctxt->ctxt);
+	comp->steps[comp->nbStep].indexExtra =
+	    xsltAllocateExtraCtxt(ctxt->ctxt);
+	comp->steps[comp->nbStep].lenExtra =
+	    xsltAllocateExtraCtxt(ctxt->ctxt);
+    } else {
+	comp->steps[comp->nbStep].previousExtra =
+	    xsltAllocateExtra(ctxt->style);
+	comp->steps[comp->nbStep].indexExtra =
+	    xsltAllocateExtra(ctxt->style);
+	comp->steps[comp->nbStep].lenExtra =
+	    xsltAllocateExtra(ctxt->style);
+    }
+    if (op == XSLT_OP_PREDICATE) {
+	comp->steps[comp->nbStep].comp = xmlXPathCompile(value);
+    }
     comp->nbStep++;
-    return(0);
+    return (0);
 }
 
 /**
@@ -264,6 +290,7 @@ xsltSwapTopCompMatch(xsltCompMatchPtr comp) {
     if (j > 0) {
 	register xmlChar *tmp;
 	register xsltOp op;
+	register xmlXPathCompExprPtr expr; 
 	i = j - 1;
 	tmp = comp->steps[i].value;
 	comp->steps[i].value = comp->steps[j].value;
@@ -274,6 +301,9 @@ xsltSwapTopCompMatch(xsltCompMatchPtr comp) {
 	op = comp->steps[i].op;
 	comp->steps[i].op = comp->steps[j].op;
 	comp->steps[j].op = op;
+	expr = comp->steps[i].comp;
+	comp->steps[i].comp = comp->steps[j].comp;
+	comp->steps[j].comp = expr;
     }
 }
 
@@ -291,6 +321,7 @@ xsltReverseCompMatch(xsltCompMatchPtr comp) {
     while (j > i) {
 	register xmlChar *tmp;
 	register xsltOp op;
+	register xmlXPathCompExprPtr expr; 
 	tmp = comp->steps[i].value;
 	comp->steps[i].value = comp->steps[j].value;
 	comp->steps[j].value = tmp;
@@ -300,27 +331,15 @@ xsltReverseCompMatch(xsltCompMatchPtr comp) {
 	op = comp->steps[i].op;
 	comp->steps[i].op = comp->steps[j].op;
 	comp->steps[j].op = op;
+	expr = comp->steps[i].comp;
+	comp->steps[i].comp = comp->steps[j].comp;
+	comp->steps[j].comp = expr;
 	j--;
 	i++;
     }
     comp->steps[comp->nbStep++].op = XSLT_OP_END;
 }
 
-/**
- * xsltCleanupCompMatch:
- * @comp:  the compiled match expression
- *
- * remove all computation state from the pattern
- */
-static void
-xsltCleanupCompMatch(xsltCompMatchPtr comp) {
-    int i;
-    
-    for (i = 0;i < comp->nbStep;i++) {
-	comp->steps[i].previous = NULL;
-    }
-}
-
 /************************************************************************
  * 									*
  * 		The interpreter for the precompiled patterns		*
@@ -391,6 +410,8 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    return(0);
 		if (step->value == NULL)
 		    continue;
+		if (step->value[0] != node->name[0])
+		    return(0);
 		if (!xmlStrEqual(step->value, node->name))
 		    return(0);
 
@@ -421,6 +442,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		if (step->value != NULL) {
 		    while (lst != NULL) {
 			if ((lst->type == XML_ELEMENT_NODE) &&
+			    (step->value[0] == lst->name[0]) &&
 			    (xmlStrEqual(step->value, lst->name)))
 			    break;
 			lst = lst->next;
@@ -433,31 +455,36 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
             case XSLT_OP_ATTR:
 		if (node->type != XML_ATTRIBUTE_NODE)
 		    return(0);
-		if (step->value == NULL)
-		    continue;
-		if (!xmlStrEqual(step->value, node->name))
-		    return(0);
-
+		if (step->value != NULL) {
+		    if (step->value[0] != node->name[0])
+			return(0);
+		    if (!xmlStrEqual(step->value, node->name))
+			return(0);
+		}
 		/* Namespace test */
 		if (node->ns == NULL) {
 		    if (step->value2 != NULL)
 			return(0);
-		} else if (node->ns->href != NULL) {
-		    if (step->value2 == NULL)
-			return(0);
+		} else if (step->value2 != NULL) {
 		    if (!xmlStrEqual(step->value2, node->ns->href))
 			return(0);
 		}
 		continue;
             case XSLT_OP_PARENT:
-		if ((node->type != XML_ELEMENT_NODE) &&
-		    (node->type != XML_ATTRIBUTE_NODE))
+		if ((node->type == XML_DOCUMENT_NODE) ||
+		    (node->type == XML_HTML_DOCUMENT_NODE) ||
+#ifdef LIBXML_DOCB_ENABLED
+		    (node->type == XML_DOCB_DOCUMENT_NODE) ||
+#endif
+		    (node->type == XML_NAMESPACE_DECL))
 		    return(0);
 		node = node->parent;
 		if (node == NULL)
 		    return(0);
 		if (step->value == NULL)
 		    continue;
+		if (step->value[0] != node->name[0])
+		    return(0);
 		if (!xmlStrEqual(step->value, node->name))
 		    return(0);
 		/* Namespace test */
@@ -485,11 +512,20 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		}
 		if (node == NULL)
 		    return(0);
+		if ((node->type == XML_DOCUMENT_NODE) ||
+		    (node->type == XML_HTML_DOCUMENT_NODE) ||
+#ifdef LIBXML_DOCB_ENABLED
+		    (node->type == XML_DOCB_DOCUMENT_NODE) ||
+#endif
+		    (node->type == XML_NAMESPACE_DECL))
+		    return(0);
 		node = node->parent;
 		while (node != NULL) {
 		    if (node == NULL)
 			return(0);
-		    if (xmlStrEqual(step->value, node->name)) {
+		    if ((node->type == XML_ELEMENT_NODE) &&
+			(step->value[0] == node->name[0]) &&
+			(xmlStrEqual(step->value, node->name))) {
 			/* Namespace test */
 			if (node->ns == NULL) {
 			    if (step->value2 == NULL)
@@ -553,6 +589,91 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		xmlNodePtr oldNode;
 		int oldCS, oldCP;
 		int pos = 0, len = 0;
+		/*
+		 * The simple existing predicate code cannot handle
+		 * properly cascaded predicates. If in this situation
+		 * compute directly the full node list once and check
+		 * if the node is in the result list.
+		 */
+		if (comp->steps[i + 1].op == XSLT_OP_PREDICATE) {
+		    xmlDocPtr prevdoc, doc;
+		    xmlXPathObjectPtr list;
+		    int index, j;
+		    int nocache = 0;
+
+		    prevdoc = (xmlDocPtr)
+			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra);
+		    index = (int)
+			XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra);
+		    list = (xmlXPathObjectPtr)
+			XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
+		    
+		    doc = node->doc;
+		    if ((list == NULL) || (prevdoc != doc)) {
+			xmlChar *query;
+			xmlXPathObjectPtr newlist;
+			xmlNodePtr parent = node->parent;
+
+			if (comp->pattern[0] == '/')
+			    query = xmlStrdup(comp->pattern);
+			else {
+			    query = xmlStrdup((const xmlChar *)"//");
+			    query = xmlStrcat(query, comp->pattern);
+			}
+			newlist = xmlXPathEval(query, ctxt->xpathCtxt);
+			xmlFree(query);
+			if (newlist == NULL)
+			    return(-1);
+			if (newlist->type != XPATH_NODESET) {
+			    xmlXPathFreeObject(newlist);
+			    return(-1);
+			}
+			
+			if ((parent == NULL) || (node->doc == NULL))
+			    nocache = 1;
+			else {
+			    while (parent->parent != NULL)
+				parent = parent->parent;
+			    if (((parent->type != XML_DOCUMENT_NODE) &&
+				 (parent->type != XML_HTML_DOCUMENT_NODE)) ||
+				 (parent != (xmlNodePtr) node->doc))
+				nocache = 1;
+			}
+			if (nocache == 0) {
+			    if (list != NULL)
+				xmlXPathFreeObject(list);
+			    list = newlist;
+
+			    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
+				(void *) list;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
+				(void *) doc;
+			    XSLT_RUNTIME_EXTRA_FREE(ctxt, select->lenExtra) =
+				(xmlFreeFunc) xmlXPathFreeObject;
+			} else
+			    list = newlist;
+		    }
+		    if ((list->nodesetval == NULL) ||
+			(list->nodesetval->nodeNr <= 0)) {
+			if (nocache == 1)
+			    xmlXPathFreeObject(list);
+			return(0);
+		    }
+		    /* TODO: store the index and use it for the scan */
+		    if (index == 0) {
+			for (j = 0;j < list->nodesetval->nodeNr;j++) {
+			    if (list->nodesetval->nodeTab[j] == node) {
+				if (nocache == 1)
+				    xmlXPathFreeObject(list);
+				return(1);
+			    }
+			}
+		    } else {
+		    }
+		    if (nocache == 1)
+			xmlXPathFreeObject(list);
+		    return(0);
+		}
 		/*
 		 * Depending on the last selection, one may need to
 		 * recompute contextSize and proximityPosition.
@@ -566,9 +687,15 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    (select->value != NULL) &&
 		    (node->type == XML_ELEMENT_NODE) &&
 		    (node->parent != NULL)) {
+		    xmlNodePtr previous;
+		    int index, nocache = 0;
 
-		    if ((select->previous != NULL) &&
-			(select->previous->parent == node->parent)) {
+		    previous = (xmlNodePtr)
+			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra);
+		    index = (int)
+			XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra);
+		    if ((previous != NULL) &&
+			(previous->parent == node->parent)) {
 			/*
 			 * just walk back to adjust the index
 			 */
@@ -576,9 +703,11 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			xmlNodePtr sibling = node;
 
 			while (sibling != NULL) {
-			    if (sibling == select->previous)
+			    if (sibling == previous)
 				break;
-			    if (xmlStrEqual(node->name, sibling->name)) {
+			    if ((node->type == XML_ELEMENT_NODE) &&
+				(node->name[0] == sibling->name[0]) &&
+				(xmlStrEqual(node->name, sibling->name))) {
 				if ((select->value2 == NULL) ||
 				    ((sibling->ns != NULL) &&
 				     (xmlStrEqual(select->value2,
@@ -592,7 +721,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    indx = 0;
 			    sibling = node;
 			    while (sibling != NULL) {
-				if (sibling == select->previous)
+				if (sibling == previous)
 				    break;
 				if ((select->value2 == NULL) ||
 				    ((sibling->ns != NULL) &&
@@ -603,10 +732,20 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    }
 			}
 			if (sibling != NULL) {
-			    pos = select->index + indx;
-			    len = select->len;
-			    select->previous = node;
-			    select->index = pos;
+			    pos = index + indx;
+			    /*
+			     * If the node is in a Value Tree we cannot
+			     * cache it !
+			     */
+			    if (node->doc != NULL) {
+				len = (int)
+				    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
+				XSLT_RUNTIME_EXTRA(ctxt,
+					select->previousExtra) = node;
+				XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				    (void *) pos;
+			    }
+			    index = pos;
 			} else
 			    pos = 0;
 		    } else {
@@ -614,14 +753,15 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			 * recompute the index
 			 */
 			xmlNodePtr siblings = node->parent->children;
+			xmlNodePtr parent = node->parent;
 
 			while (siblings != NULL) {
 			    if (siblings->type == XML_ELEMENT_NODE) {
 				if (siblings == node) {
 				    len++;
 				    pos = len;
-				} else if (xmlStrEqual(node->name,
-					   siblings->name)) {
+				} else if ((node->name[0] == siblings->name[0])
+			       && (xmlStrEqual(node->name, siblings->name))) {
 				    if ((select->value2 == NULL) ||
 					((siblings->ns != NULL) &&
 					 (xmlStrEqual(select->value2,
@@ -631,18 +771,44 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    }
 			    siblings = siblings->next;
 			}
+			if ((parent == NULL) || (node->doc == NULL))
+			    nocache = 1;
+			else {
+			    while (parent->parent != NULL)
+				parent = parent->parent;
+			    if (((parent->type != XML_DOCUMENT_NODE) &&
+				 (parent->type != XML_HTML_DOCUMENT_NODE)) ||
+				 (parent != (xmlNodePtr) node->doc))
+				nocache = 1;
+			}
 		    }
 		    if (pos != 0) {
 			ctxt->xpathCtxt->contextSize = len;
 			ctxt->xpathCtxt->proximityPosition = pos;
-			select->previous = node;
-			select->index = pos;
-			select->len = len;
+			/*
+			 * If the node is in a Value Tree we cannot
+			 * cache it !
+			 */
+			if ((node->doc != NULL) && (nocache == 0)) {
+			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
+				node;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				(void *) pos;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
+				(void *) len;
+			}
 		    }
 		} else if ((select != NULL) && (select->op == XSLT_OP_ALL) &&
 			   (node->type == XML_ELEMENT_NODE)) {
-		    if ((select->previous != NULL) &&
-			(select->previous->parent == node->parent)) {
+		    xmlNodePtr previous;
+		    int index, nocache = 0;
+
+		    previous = (xmlNodePtr)
+			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra);
+		    index = (int)
+			XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra);
+		    if ((previous != NULL) &&
+			(previous->parent == node->parent)) {
 			/*
 			 * just walk back to adjust the index
 			 */
@@ -650,7 +816,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			xmlNodePtr sibling = node;
 
 			while (sibling != NULL) {
-			    if (sibling == select->previous)
+			    if (sibling == previous)
 				break;
 			    if (sibling->type == XML_ELEMENT_NODE)
 				indx++;
@@ -661,7 +827,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    indx = 0;
 			    sibling = node;
 			    while (sibling != NULL) {
-				if (sibling == select->previous)
+				if (sibling == previous)
 				    break;
 				if (sibling->type == XML_ELEMENT_NODE)
 				    indx--;
@@ -669,10 +835,19 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    }
 			}
 			if (sibling != NULL) {
-			    pos = select->index + indx;
-			    len = select->len;
-			    select->previous = node;
-			    select->index = pos;
+			    pos = index + indx;
+			    /*
+			     * If the node is in a Value Tree we cannot
+			     * cache it !
+			     */
+			    if (node->doc != NULL) {
+				len = (int)
+				    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
+				XSLT_RUNTIME_EXTRA(ctxt,
+					select->previousExtra) = node;
+				XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				    (void *) pos;
+			    }
 			} else
 			    pos = 0;
 		    } else {
@@ -680,6 +855,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			 * recompute the index
 			 */
 			xmlNodePtr siblings = node->parent->children;
+			xmlNodePtr parent = node->parent;
 
 			while (siblings != NULL) {
 			    if (siblings->type == XML_ELEMENT_NODE) {
@@ -690,13 +866,32 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    }
 			    siblings = siblings->next;
 			}
+			if ((parent == NULL) || (node->doc == NULL))
+			    nocache = 1;
+			else {
+			    while (parent->parent != NULL)
+				parent = parent->parent;
+			    if (((parent->type != XML_DOCUMENT_NODE) &&
+				 (parent->type != XML_HTML_DOCUMENT_NODE)) ||
+				 (parent != (xmlNodePtr) node->doc))
+				nocache = 1;
+			}
 		    }
 		    if (pos != 0) {
 			ctxt->xpathCtxt->contextSize = len;
 			ctxt->xpathCtxt->proximityPosition = pos;
-			select->previous = node;
-			select->index = pos;
-			select->len = len;
+			/*
+			 * If the node is in a Value Tree we cannot
+			 * cache it !
+			 */
+			if ((node->doc != NULL) && (nocache == 0)) {
+			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
+				node;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				(void *) pos;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
+				(void *) len;
+			}
 		    }
 		}
 		oldNode = ctxt->node;
@@ -704,22 +899,9 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 
 		if (step->value == NULL)
 		    goto wrong_index;
+		if (step->comp == NULL)
+		    goto wrong_index;
 
-		if (step->comp == NULL) {
-		    step->comp = xmlXPathCompile(step->value);
-		    if (step->comp == NULL)
-			goto wrong_index;
-		}
-		if (comp->nsList == NULL) {
-		    int j = 0;
-
-		    comp->nsList = xmlGetNsList(node->doc, node);
-		    if (comp->nsList != NULL) {
-			while (comp->nsList[j] != NULL)
-			    j++;
-		    }
-		    comp->nsNr = j;
-		}
 		if (!xsltEvalXPathPredicate(ctxt, step->comp, comp->nsList,
 			                    comp->nsNr))
 		    goto wrong_index;
@@ -823,7 +1005,7 @@ xsltTestCompMatchList(xsltTransformContextPtr ctxt, xmlNodePtr node,
 
 
 #define PUSH(op, val, val2) 						\
-    if (xsltCompMatchAdd(ctxt->comp, (op), (val), (val2))) goto error;
+    if (xsltCompMatchAdd(ctxt, ctxt->comp, (op), (val), (val2))) goto error;
 
 #define SWAP() 						\
     xsltSwapTopCompMatch(ctxt->comp);
@@ -850,36 +1032,43 @@ xsltTestCompMatchList(xsltTransformContextPtr ctxt, xmlNodePtr node,
 
 static xmlChar *
 xsltScanLiteral(xsltParserContextPtr ctxt) {
-    const xmlChar *q;
+    const xmlChar *q, *cur;
     xmlChar *ret = NULL;
+    int val, len;
 
     SKIP_BLANKS;
     if (CUR == '"') {
         NEXT;
-	q = CUR_PTR;
-	while ((IS_CHAR(CUR)) && (CUR != '"'))
-	    NEXT;
-	if (!IS_CHAR(CUR)) {
-	    /* XP_ERROR(XPATH_UNFINISHED_LITERAL_ERROR); */
+	cur = q = CUR_PTR;
+	val = xmlStringCurrentChar(NULL, cur, &len);
+	while ((IS_CHAR(val)) && (val != '"')) {
+	    cur += len;
+	    val = xmlStringCurrentChar(NULL, cur, &len);
+	}
+	if (!IS_CHAR(val)) {
 	    ctxt->error = 1;
 	    return(NULL);
 	} else {
-	    ret = xmlStrndup(q, CUR_PTR - q);
-	    NEXT;
+	    ret = xmlStrndup(q, cur - q);
         }
+	cur += len;
+	CUR_PTR = cur;
     } else if (CUR == '\'') {
         NEXT;
-	q = CUR_PTR;
-	while ((IS_CHAR(CUR)) && (CUR != '\''))
-	    NEXT;
-	if (!IS_CHAR(CUR)) {
-	    /* XP_ERROR(XPATH_UNFINISHED_LITERAL_ERROR); */
+	cur = q = CUR_PTR;
+	val = xmlStringCurrentChar(NULL, cur, &len);
+	while ((IS_CHAR(val)) && (val != '\'')) {
+	    cur += len;
+	    val = xmlStringCurrentChar(NULL, cur, &len);
+	}
+	if (!IS_CHAR(val)) {
 	    ctxt->error = 1;
 	    return(NULL);
 	} else {
-	    ret = xmlStrndup(q, CUR_PTR - q);
-	    NEXT;
+	    ret = xmlStrndup(q, cur - q);
         }
+	cur += len;
+	CUR_PTR = cur;
     } else {
 	/* XP_ERROR(XPATH_START_LITERAL_ERROR); */
 	ctxt->error = 1;
@@ -904,36 +1093,28 @@ xsltScanLiteral(xsltParserContextPtr ctxt) {
 
 static xmlChar *
 xsltScanName(xsltParserContextPtr ctxt) {
-    xmlChar buf[XML_MAX_NAMELEN];
-    int len = 0;
+    const xmlChar *q, *cur;
+    xmlChar *ret = NULL;
+    int val, len;
 
     SKIP_BLANKS;
-    if (!IS_LETTER(CUR) && (CUR != '_') &&
-        (CUR != ':')) {
-	return(NULL);
-    }
 
-    while ((IS_LETTER(NXT(len))) || (IS_DIGIT(NXT(len))) ||
-           (NXT(len) == '.') || (NXT(len) == '-') ||
-	   (NXT(len) == '_') || 
-	   (IS_COMBINING(NXT(len))) ||
-	   (IS_EXTENDER(NXT(len)))) {
-	buf[len] = NXT(len);
-	len++;
-	if (len >= XML_MAX_NAMELEN) {
-	    xmlGenericError(xmlGenericErrorContext, 
-	       "xsltScanName: reached XML_MAX_NAMELEN limit\n");
-	    while ((IS_LETTER(NXT(len))) || (IS_DIGIT(NXT(len))) ||
-		   (NXT(len) == '.') || (NXT(len) == '-') ||
-		   (NXT(len) == '_') || (NXT(len) == ':') || 
-		   (IS_COMBINING(NXT(len))) ||
-		   (IS_EXTENDER(NXT(len))))
-		 len++;
-	    break;
-	}
+    cur = q = CUR_PTR;
+    val = xmlStringCurrentChar(NULL, cur, &len);
+    if (!IS_LETTER(val) && (val != '_') && (val != ':'))
+	return(NULL);
+
+    while ((IS_LETTER(val)) || (IS_DIGIT(val)) ||
+           (val == '.') || (val == '-') ||
+	   (val == '_') || 
+	   (IS_COMBINING(val)) ||
+	   (IS_EXTENDER(val))) {
+	cur += len;
+	val = xmlStringCurrentChar(NULL, cur, &len);
     }
-    SKIP(len);
-    return(xmlStrndup(buf, len));
+    ret = xmlStrndup(q, cur - q);
+    CUR_PTR = cur;
+    return(ret);
 }
 
 /**
@@ -947,35 +1128,28 @@ xsltScanName(xsltParserContextPtr ctxt) {
 
 static xmlChar *
 xsltScanNCName(xsltParserContextPtr ctxt) {
-    xmlChar buf[XML_MAX_NAMELEN];
-    int len = 0;
+    const xmlChar *q, *cur;
+    xmlChar *ret = NULL;
+    int val, len;
 
     SKIP_BLANKS;
-    if (!IS_LETTER(CUR) && (CUR != '_')) {
-	return(NULL);
-    }
 
-    while ((IS_LETTER(NXT(len))) || (IS_DIGIT(NXT(len))) ||
-           (NXT(len) == '.') || (NXT(len) == '-') ||
-	   (NXT(len) == '_') ||
-	   (IS_COMBINING(NXT(len))) ||
-	   (IS_EXTENDER(NXT(len)))) {
-	buf[len] = NXT(len);
-	len++;
-	if (len >= XML_MAX_NAMELEN) {
-	    xmlGenericError(xmlGenericErrorContext, 
-	       "xsltScanNCName: reached XML_MAX_NAMELEN limit\n");
-	    while ((IS_LETTER(NXT(len))) || (IS_DIGIT(NXT(len))) ||
-		   (NXT(len) == '.') || (NXT(len) == '-') ||
-		   (NXT(len) == '_') ||
-		   (IS_COMBINING(NXT(len))) ||
-		   (IS_EXTENDER(NXT(len))))
-		 len++;
-	    break;
-	}
+    cur = q = CUR_PTR;
+    val = xmlStringCurrentChar(NULL, cur, &len);
+    if (!IS_LETTER(val) && (val != '_'))
+	return(NULL);
+
+    while ((IS_LETTER(val)) || (IS_DIGIT(val)) ||
+           (val == '.') || (val == '-') ||
+	   (val == '_') ||
+	   (IS_COMBINING(val)) ||
+	   (IS_EXTENDER(val))) {
+	cur += len;
+	val = xmlStringCurrentChar(NULL, cur, &len);
     }
-    SKIP(len);
-    return(xmlStrndup(buf, len));
+    ret = xmlStrndup(q, cur - q);
+    CUR_PTR = cur;
+    return(ret);
 }
 
 /**
@@ -1214,7 +1388,7 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 	    goto error;
 	}
 	PUSH(XSLT_OP_ATTR, token, URL);
-	return;
+	goto parse_predicate;
     }
     if (token == NULL)
 	token = xsltScanName(ctxt);
@@ -1280,11 +1454,17 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 		xmlFree(token);
 		token = xsltScanName(ctxt);
 		if (token == NULL) {
-		    xsltPrintErrorContext(NULL, NULL, NULL); /* TODO */
-		    xsltGenericError(xsltGenericErrorContext,
+	            if (CUR == '*') {
+            	        NEXT;
+	                PUSH(XSLT_OP_ALL, token, NULL);
+	                goto parse_predicate;
+	            } else {
+		        xsltPrintErrorContext(NULL, NULL, NULL); /* TODO */
+		        xsltGenericError(xsltGenericErrorContext,
 			    "xsltCompileStepPattern : QName expected\n");
-		    ctxt->error = 1;
-		    goto error;
+		        ctxt->error = 1;
+		        goto error;
+		    }
 		}
 		URI = xsltGetQNameURI(ctxt->elem, &token);
 		if (token == NULL) {
@@ -1348,19 +1528,26 @@ parse_predicate:
 	level++;
 	NEXT;
 	q = CUR_PTR;
-	/* TODO: avoid breaking in strings ... */
-	while (IS_CHAR(CUR)) {
+	while (CUR != 0) {
 	    /* Skip over nested predicates */
 	    if (CUR == '[')
 		level++;
-	    if (CUR == ']') {
+	    else if (CUR == ']') {
 		level--;
 		if (level == 0)
 		    break;
+	    } else if (CUR == '"') {
+		NEXT;
+		while ((CUR != 0) && (CUR != '"'))
+		    NEXT;
+	    } else if (CUR == '\'') {
+		NEXT;
+		while ((CUR != 0) && (CUR != '\''))
+		    NEXT;
 	    }
 	    NEXT;
 	}
-	if (!IS_CHAR(CUR)) {
+	if (CUR == 0) {
 	    xsltPrintErrorContext(NULL, NULL, NULL); /* TODO */
 	    xsltGenericError(xsltGenericErrorContext,
 		    "xsltCompileStepPattern : ']' expected\n");
@@ -1500,6 +1687,8 @@ error:
  * @pattern: an XSLT pattern
  * @doc:  the containing document
  * @node:  the containing element
+ * @style:  the stylesheet
+ * @runtime:  the transformation context, if done at run-time
  *
  * Compile the XSLT pattern and generates a list of precompiled form suitable
  * for fast matching.
@@ -1510,10 +1699,12 @@ error:
  */
 
 xsltCompMatchPtr
-xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
+xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc,
+	           xmlNodePtr node, xsltStylesheetPtr style,
+		   xsltTransformContextPtr runtime) {
     xsltParserContextPtr ctxt = NULL;
     xsltCompMatchPtr element, first = NULL, previous = NULL;
-    int current, start, end;
+    int current, start, end, level, j;
 
     if (pattern == NULL) {
 	xsltPrintErrorContext(NULL, NULL, node); /* TODO */
@@ -1522,7 +1713,7 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
 	return(NULL);
     }
 
-    ctxt = xsltNewParserContext();
+    ctxt = xsltNewParserContext(style, runtime);
     if (ctxt == NULL)
 	return(NULL);
     ctxt->doc = doc;
@@ -1533,8 +1724,23 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
 	while (IS_BLANK(pattern[current]))
 	    current++;
 	end = current;
-	while ((pattern[end] != 0) && (pattern[end] != '|'))
+	level = 0;
+	while ((pattern[end] != 0) && ((pattern[end] != '|') || (level != 0))) {
+	    if (pattern[end] == '[')
+		level++;
+	    else if (pattern[end] == ']')
+		level--;
+	    else if (pattern[end] == '\'') {
+		end++;
+		while ((pattern[end] != 0) && (pattern[end] != '\''))
+		    end++;
+	    } else if (pattern[end] == '"') {
+		end++;
+		while ((pattern[end] != 0) && (pattern[end] != '"'))
+		    end++;
+	    }
 	    end++;
+	}
 	if (current == end) {
 	    xsltPrintErrorContext(NULL, NULL, node); /* TODO */
 	    xsltGenericError(xsltGenericErrorContext,
@@ -1557,6 +1763,14 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
 	    goto error;
 	ctxt->cur = &(ctxt->base)[current - start];
 	element->pattern = ctxt->base;
+	element->nsList = xmlGetNsList(doc, node);
+	j = 0;
+	if (element->nsList != NULL) {
+	    while (element->nsList[j] != NULL)
+		j++;
+	}
+	element->nsNr = j;
+
 
 #ifdef WITH_XSLT_DEBUG_PATTERN
 	xsltGenericDebug(xsltGenericDebugContext,
@@ -1564,8 +1778,14 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
 			 element->pattern);
 #endif
 	xsltCompileLocationPathPattern(ctxt);
-	if (ctxt->error)
+	if (ctxt->error) {
+	    xsltPrintErrorContext(NULL, style, node);
+	    xsltGenericError(xsltGenericErrorContext,
+			     "xsltCompilePattern : failed to compile '%s'\n",
+			     element->pattern);
+	    style->errors++;
 	    goto error;
+	}
 
 	/*
 	 * Reverse for faster interpretation.
@@ -1625,6 +1845,7 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
 	xsltPrintErrorContext(NULL, NULL, node); /* TODO */
         xsltGenericError(xsltGenericErrorContext,
 			 "xsltCompilePattern : NULL pattern\n");
+	style->errors++;
 	goto error;
     }
 
@@ -1667,7 +1888,7 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	return(-1);
 
     priority = cur->priority;
-    pat = xsltCompilePattern(cur->match, style->doc, cur->elem);
+    pat = xsltCompilePattern(cur->match, style->doc, cur->elem, style, NULL);
     while (pat) {
 	next = pat->next;
 	pat->next = NULL;
@@ -1691,7 +1912,6 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	    else
 		top = (xsltCompMatchPtr *) &(style->attrMatch);
 	    break;
-        case XSLT_OP_ELEM:
         case XSLT_OP_CHILD:
         case XSLT_OP_PARENT:
         case XSLT_OP_ANCESTOR:
@@ -1732,12 +1952,12 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	case XSLT_OP_TEXT:
 	    top = (xsltCompMatchPtr *) &(style->textMatch);
 	    break;
+        case XSLT_OP_ELEM:
 	case XSLT_OP_NODE:
 	    if (pat->steps[0].value != NULL)
 		name = pat->steps[0].value;
 	    else
 		top = (xsltCompMatchPtr *) &(style->elemMatch);
-	    
 	    break;
 	}
 	if (name != NULL) {
@@ -2000,13 +2220,7 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
  * the ones it imports.
  */
 void
-xsltCleanupTemplates(xsltStylesheetPtr style) {
-    while (style != NULL) {
-	xmlHashScan((xmlHashTablePtr) style->templatesHash,
-		    (xmlHashScanner) xsltCleanupCompMatch, NULL);
-
-	style = xsltNextImport(style);
-    }
+xsltCleanupTemplates(xsltStylesheetPtr style ATTRIBUTE_UNUSED) {
 }
 
 /**
@@ -2043,19 +2257,23 @@ xsltFreeTemplateHashes(xsltStylesheetPtr style) {
  * xsltMatchPattern
  * @node: a node in the source tree
  * @pattern: an XSLT pattern
+ * @ctxtdoc:  context document (for namespaces)
+ * @ctxtnode:  context node (for namespaces)
  *
  * Determine if a node matches a pattern.
  */
 int
 xsltMatchPattern(xsltTransformContextPtr context,
 		 xmlNodePtr node,
-		 const xmlChar *pattern)
+		 const xmlChar *pattern,
+		 xmlDocPtr ctxtdoc,
+		 xmlNodePtr ctxtnode)
 {
     int match = 0;
     xsltCompMatchPtr first, comp;
 
     if ((context != NULL) && (pattern != NULL)) {
-	first = xsltCompilePattern(pattern);
+	first = xsltCompilePattern(pattern, ctxtdoc, ctxtnode);
 	for (comp = first; comp != NULL; comp = comp->next) {
 	    match = xsltTestCompMatch(context, comp, node, NULL, NULL);
 	    if (match)
