@@ -19,6 +19,12 @@ extern "C" {
 /*
  * Interfaces
  */
+extern const xmlChar *xsltExtMarker;
+
+xsltElemPreCompPtr xsltDocumentComp	(xsltStylesheetPtr style,
+					 xmlNodePtr inst,
+					 xsltTransformFunction function);
+
 void		xsltStylePreCompute	(xsltStylesheetPtr style,
 					 xmlNodePtr inst);
 void		xsltFreeStylePreComps	(xsltStylesheetPtr style);
