@@ -88,6 +88,7 @@ xmlParserInputPtr xmlNoNetExternalEntityLoader(const char *URL,
 static int debug = 0;
 static int repeat = 0;
 static int timing = 0;
+static int dumpextensions = 0;
 static int novalid = 0;
 static int noout = 0;
 #ifdef LIBXML_DOCB_ENABLED
@@ -236,6 +237,7 @@ static void
 xsltProcess(xmlDocPtr doc, xsltStylesheetPtr cur, const char *filename) {
     xmlDocPtr res;
     xsltTransformContextPtr ctxt;
+   
 
 #ifdef LIBXML_XINCLUDE_ENABLED
     if (xinclude) {
@@ -282,6 +284,8 @@ xsltProcess(xmlDocPtr doc, xsltStylesheetPtr cur, const char *filename) {
 	}
 	if (ctxt->state == XSLT_STATE_ERROR)
 	    errorno = 9;
+	if (ctxt->state == XSLT_STATE_STOPPED)
+	    errorno = 10;
 	xsltFreeTransformContext(ctxt);
 	if (timing) {
 	    if (repeat)
@@ -361,6 +365,7 @@ static void usage(const char *name) {
     printf("\t--timing: display the time used\n");
     printf("\t--repeat: run the transformation 20 times\n");
     printf("\t--debug: dump the tree of the result instead\n");
+    printf("\t--dumpextensions: dump the registered extension elements and functions to stdout\n");
     printf("\t--novalid skip the Dtd loading phase\n");
     printf("\t--noout: do not dump the result\n");
     printf("\t--maxdepth val : increase the maximum depth\n");
@@ -540,7 +545,11 @@ main(int argc, char **argv)
                 if (value > 0)
                     xsltMaxDepth = value;
             }
-        } else {
+        } else if ((!strcmp(argv[i],"-dumpextensions"))||
+			(!strcmp(argv[i],"--dumpextensions"))) {
+		dumpextensions++;
+		
+	} else {
             fprintf(stderr, "Unknown option %s\n", argv[i]);
             usage(argv[0]);
             return (3);
@@ -565,6 +574,9 @@ main(int argc, char **argv)
     exsltRegisterAll();
     xsltRegisterTestModule();
 
+    if (dumpextensions) 
+	xsltDebugDumpExtensions(NULL);
+
     for (i = 1; i < argc; i++) {
         if ((!strcmp(argv[i], "-maxdepth")) ||
             (!strcmp(argv[i], "--maxdepth"))) {
@@ -605,6 +617,10 @@ main(int argc, char **argv)
 		}
 		cur = xsltParseStylesheetDoc(style);
 		if (cur != NULL) {
+		    if (cur->errors != 0) {
+			errorno = 5;
+			goto done;
+		    }
 		    if (cur->indent == 1)
 			xmlIndentTreeOutput = 1;
 		    else
